##!/bin/bash

# Base path ot installation: /tmp or HOME dir
if [ -z "$HOME" ]; then 
	PREFIX=/tmp
	else
	PREFIX=$HOME
fi

########## Setting parameters ##########

# Parameters 
# 1 - intallation path
# 2 - port
# 3 - sib db folder
# 4 - dbus sib file name
# 5 - sib-tcp pid file name

# installation directory, default <$PREFIX>/inst
if [ $# -ge 1 ]; then
	INST_PATH=$1;
	else
	INST_PATH=$PREFIX"/inst"
fi

# sib-tcp port, default 10000 + RANDOM_NUM(1-1000)  
if [ $# -ge 2 ]; then
	PORT=$2;
	else
	PORT=$[ 10000+ ( $RANDOM % 1000 )  + 1 ]
fi

# sib db path, default "." - current directory 
if [ $# -ge 3 ]; then
	SIB_DB_FOLDER=$3;
	else
	SIB_DB_FOLDER="."
fi

# dbus sib file name, default "dbus-sib_<$USER>_<RANDOM_NUM>.pid" 
if [ $# -ge 4 ]; then
	SIB_DBUS_FILE=$4;
	else
	SIB_DBUS_FILE=dbus-sib_"$USER"_$[ ( $RANDOM % 1000 )  + 1 ]
fi


# sib-tcp pid file name, default "sib_tcp_cc_id_<$USER>_<RANDOM_NUM>.pid" 
if [ $# -ge 5 ]; then
	SIB_TCP_PID_FILE=$5;
	else
	SIB_TCP_PID_FILE=sib_tcp_cc_id_"$USER"_$[ ( $RANDOM % 1000 )  + 1 ]
fi



########### Initialization #############

INIT_DIR=`pwd`
CONFIG_FILE=$INIT_DIR/"config.log"
INSTALLATION_FILE=$INIT_DIR/"install.log"
SIB_PKGC_PATH=$INST_PATH"/lib/pkgconfig/"
SIB_PKGC64_PATH=$INST_PATH"/lib64/pkgconfig/"

export SIB_PORT_DFLT=$PORT
export SIB_DB_FOLDER=$SIB_DB_FOLDER
export SIB_TCP_PID_FILE=$SIB_TCP_PID_FILE
export SIB_DBUS_FILE=$SIB_DBUS_FILE
export PKG_CONFIG_PATH=$PKG_CONFIG_PATH:$SIB_PKGC_PATH:$SIB_PKGC64_PATH


#Paths to modules
PATHS=( "./sources/libwhiteboard-2.3/"  \
        "./sources/sib-tcp-0.8/" \
        "./sources/redsibd-0.4.1/" )

#Modules names
MODULE_NAMES=( "WHITEBOARD" 
 #       "WHITEBOARD DEAMON" 
#        "WHITEBOARD-SIB-TCP" 
        "SIB-TCP" 
        "REDLAND SIB" )

PATHS_COUNT=${#PATHS[@]}



######### Information output ###########
echo ""
echo "--- Installation information ---"
echo " - installation path:" $INST_PATH
echo " - SIB port:" $PORT
echo " - SIB-TCP PID file: '"$SIB_TCP_PID_FILE"'"
echo " - DBUS-SIB file: '"$SIB_DBUS_FILE"'"
echo " - SIB DB FOLDER: '"$SIB_DB_FOLDER"'"
echo ""
echo "Config information will be in the config file:"
echo $CONFIG_FILE
echo ""
echo "Installation log will be in the file:"
echo $INSTALLATION_FILE
echo ""
echo "--- Script parameters ---"
echo "Using script parameters you can change: "
echo " 1 - intallation path;"
echo " 2 - port;"
echo " 3 - Database SIB folder;"
echo " 4 - Dbus SIB file name;"
echo " 5 - SIB-TCP Pid file name."
echo "For example:"
echo "./install.sh /home/student/testsib 10999 /home/student/temp/ dbus_sib_student sib_tcp_student_pid_file"
echo ""
echo ""
read -p "Press [Enter] key to start installation or CTRL+C to cancel..."



############ Installation ##############

echo ""
echo "--- Installation... ---"
echo "=== "`date`" ===" > $INSTALLATION_FILE

for (( i=0; i<$PATHS_COUNT; i++ ));
do
	echo ""
	echo " --- "${MODULE_NAMES[$i]}" --- "
	echo "Path: "${PATHS[$i]}
    echo "Installing, please wait..."
	sleep 1
	cd ${PATHS[$i]} 
	echo "" >> $INSTALLATION_FILE
	echo " --- "${MODULE_NAMES[$i]}" --- " >> $INSTALLATION_FILE
	#./autogen.sh --with-debug --prefix=$INST_PATH
	./configure --with-debug --prefix=$INST_PATH > /dev/null  2>>$INSTALLATION_FILE
	make clean > /dev/null 2>/dev/null
	make install > /dev/null 2>>$INSTALLATION_FILE
		
	if [ $? == 0 ]; then
	    echo " --- Installed successfully! ---"
	    echo " --- Installed successfully! ---" >> $INSTALLATION_FILE
	else
	    echo " --- Installation failed! ---"
	    echo " --- Installation failed! ---" >> $INSTALLATION_FILE
	    exit 1
	fi
	            
	
    cd $INIT_DIR
done

echo ""
echo "!!!Happy end!!!"
echo ""

echo "Local SIB configuration" > $CONFIG_FILE
echo " - installation path:" $INST_PATH >> $CONFIG_FILE
echo " - SIB port:" $PORT >> $CONFIG_FILE
echo " - SIB-TCP PID file: '"$SIB_TCP_PID_FILE"'" >> $CONFIG_FILE
echo " - SIB DB FOLDER: '"$SIB_DB_FOLDER"'" >> $CONFIG_FILE
echo " - DBUS-SIB file: '"$SIB_DBUS_FILE"'" >> $CONFIG_FILE


echo
cat  $CONFIG_FILE 
echo

echo "export PKG_CONFIG_PATH=\$PKG_CONFIG_PATH:$INST_PATH/lib64/pkgconfig" >> ~/.bashrc
echo "export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$INST_PATH/lib64" >> ~/.bashrc
